@@ -16,10 +16,12 @@
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
 import datetime
+import hashlib
 import hmac
-import socket
 import json
-import urllib
+import urllib.request
+import urllib.parse
+import urllib.error
 import time
 import errno
 import socket
@@ -27,14 +29,10 @@ from socket import gaierror
 from sys import exc_info
 from ssl import SSLError, cert_time_to_seconds
 from threading import Lock, BoundedSemaphore, Thread, Event, currentThread
-
-import six
-
 import offlineimap.accounts
 from offlineimap import imaplibutil, imaputil, threadutil, OfflineImapError
 from offlineimap.ui import getglobalui
 
-
 try:
     import gssapi
     have_gss = True
@@ -42,7 +40,7 @@ except ImportError:
     have_gss = False
 
 
-class IMAPServer(object):
+class IMAPServer:
     """Initializes all variables from an IMAPRepository() instance
 
     Various functions, such as acquireconnection() return an IMAP4
@@ -62,13 +60,13 @@ class IMAPServer(object):
         self.preauth_tunnel = repos.getpreauthtunnel()
         self.transport_tunnel = repos.gettransporttunnel()
         if self.preauth_tunnel and self.transport_tunnel:
-            raise OfflineImapError('%s: '% repos +
-                'you must enable precisely one '
-                'type of tunnel (preauth or transport), '
-                'not both', OfflineImapError.ERROR.REPO)
+            raise OfflineImapError('%s: ' % repos +
+                                   'you must enable precisely one '
+                                   'type of tunnel (preauth or transport), '
+                                   'not both', OfflineImapError.ERROR.REPO)
         self.tunnel = \
             self.preauth_tunnel if self.preauth_tunnel \
-            else self.transport_tunnel
+                else self.transport_tunnel
 
         self.username = \
             None if self.preauth_tunnel else repos.getuser()
@@ -94,18 +92,18 @@ class IMAPServer(object):
         self.sslclientkey = repos.getsslclientkey()
         self.sslcacertfile = repos.getsslcacertfile()
         if self.sslcacertfile is None:
-            self.__verifycert = None # Disable cert verification.
-                                     # This way of working sucks hard...
+            self.__verifycert = None  # Disable cert verification.
+            # This way of working sucks hard...
         self.fingerprint = repos.get_ssl_fingerprint()
         self.tlslevel = repos.gettlslevel()
         self.sslversion = repos.getsslversion()
         self.starttls = repos.getstarttls()
 
         if self.usessl \
-           and self.tlslevel is not "tls_compat" \
-           and self.sslversion is None:
+                and self.tlslevel != "tls_compat" \
+                and self.sslversion is None:
             raise Exception("When 'tls_level' is not 'tls_compat' "
-                "the 'ssl_version' must be set explicitly.")
+                            "the 'ssl_version' must be set explicitly.")
 
         self.oauth2_refresh_token = repos.getoauth2_refresh_token()
         self.oauth2_access_token = repos.getoauth2_access_token()
@@ -164,32 +162,34 @@ class IMAPServer(object):
             self.ui.warn("PySocks not installed, ignoring proxy option.")
         except (AttributeError, ValueError) as e:
             self.ui.warn("Bad proxy option %s for account %s: %s "
-                "Ignoring %s option."%
-                (proxy, self.repos.account.name, e, proxysection))
+                         "Ignoring %s option." %
+                         (proxy, self.repos.account.name, e, proxysection))
         return dfltsocket
 
     def __getpassword(self):
         """Returns the server password or None"""
 
-        if self.goodpassword != None: # use cached good one first
+        if self.goodpassword is not None:  # use cached good one first
             return self.goodpassword
 
-        if self.password != None and self.passworderror == None:
-            return self.password # non-failed preconfigured one
+        if self.password is not None and self.passworderror is None:
+            return self.password  # non-failed preconfigured one
 
         # get 1) configured password first 2) fall back to asking via UI
         self.password = self.repos.getpassword() or \
-            self.ui.getpass(self.username, self.config, self.passworderror)
+                        self.ui.getpass(self.username, self.config, self.passworderror)
         self.passworderror = None
         return self.password
 
     def __md5handler(self, response):
         challenge = response.strip()
-        self.ui.debug('imap', '__md5handler: got challenge %s'% challenge)
+        self.ui.debug('imap', '__md5handler: got challenge %s' % challenge)
 
         passwd = self.__getpassword()
-        retval = self.username + ' ' + hmac.new(passwd, challenge).hexdigest()
-        self.ui.debug('imap', '__md5handler: returning %s'% retval)
+        retval = self.username + ' ' +\
+                 hmac.new(bytes(passwd, encoding='utf-8'), challenge,
+                          digestmod=hashlib.md5).hexdigest()
+        self.ui.debug('imap', '__md5handler: returning %s' % retval)
         return retval
 
     def __loginauth(self, imapobj):
@@ -205,19 +205,18 @@ class IMAPServer(object):
         authc = self.username
         if not authc:
             raise OfflineImapError("No username provided for '%s'"
-                                    % self.repos.getname(),
+                                   % self.repos.getname(),
                                    OfflineImapError.ERROR.REPO)
 
         passwd = self.__getpassword()
-        authz = b''
-        if self.user_identity != None:
+        authz = ''
+        NULL = '\x00'
+        if self.user_identity is not None:
             authz = self.user_identity
-        # At this point all authz, authc and passwd are expected bytes encoded
-        # in UTF-8.
-        NULL = b'\x00'
+
         retval = NULL.join((authz, authc, passwd))
-        logsafe_retval = NULL.join((authz, authc, b'(passwd hidden for log)'))
-        self.ui.debug('imap', '__plainhandler: returning %s'% logsafe_retval)
+        self.ui.debug('imap', '__plainhandler: returning %s %s '
+                      '(passwd hidden for log)' % (authz, authc))
         return retval
 
     def __xoauth2handler(self, response):
@@ -230,8 +229,8 @@ class IMAPServer(object):
         if self.oauth2_access_token is None:
             if self.oauth2_request_url is None:
                 raise OfflineImapError("No remote oauth2_request_url for "
-                    "repository '%s' specified."%
-                    self, OfflineImapError.ERROR.REPO)
+                                       "repository '%s' specified." %
+                                       self, OfflineImapError.ERROR.REPO)
 
             # Generate new access token.
             params = {}
@@ -240,41 +239,43 @@ class IMAPServer(object):
             params['refresh_token'] = self.oauth2_refresh_token
             params['grant_type'] = 'refresh_token'
 
-            self.ui.debug('imap', 'xoauth2handler: url "%s"'%
-                self.oauth2_request_url)
-            self.ui.debug('imap', 'xoauth2handler: params "%s"'% params)
+            self.ui.debug('imap', 'xoauth2handler: url "%s"' %
+                          self.oauth2_request_url)
+            self.ui.debug('imap', 'xoauth2handler: params "%s"' % params)
 
             original_socket = socket.socket
             socket.socket = self.authproxied_socket
             try:
-                response = urllib.urlopen(
-                    self.oauth2_request_url, urllib.urlencode(params)).read()
+                response = urllib.request.urlopen(
+                    self.oauth2_request_url, urllib.parse.urlencode(params).encode('utf-8')).read()
             except Exception as e:
                 try:
-                    msg = "%s (configuration is: %s)"% (e, str(params))
+                    msg = "%s (configuration is: %s)" % (e, str(params))
                 except Exception as eparams:
-                    msg = "%s [cannot display configuration: %s]"% (e, eparams)
-                six.reraise(type(e), type(e)(msg), exc_info()[2])
+                    msg = "%s [cannot display configuration: %s]" % (e, eparams)
+
+                self.ui.error(msg)
+                raise
             finally:
                 socket.socket = original_socket
 
             resp = json.loads(response)
-            self.ui.debug('imap', 'xoauth2handler: response "%s"'% resp)
-            if u'error' in resp:
-                raise OfflineImapError("xoauth2handler got: %s"% resp,
-                    OfflineImapError.ERROR.REPO)
+            self.ui.debug('imap', 'xoauth2handler: response "%s"' % resp)
+            if 'error' in resp:
+                raise OfflineImapError("xoauth2handler got: %s" % resp,
+                                       OfflineImapError.ERROR.REPO)
             self.oauth2_access_token = resp['access_token']
-            if u'expires_in' in resp:
+            if 'expires_in' in resp:
                 self.oauth2_access_token_expires_at = now + datetime.timedelta(
-                    seconds=resp['expires_in']/2
+                    seconds=resp['expires_in'] / 2
                 )
 
-        self.ui.debug('imap', 'xoauth2handler: access_token "%s expires %s"'% (
+        self.ui.debug('imap', 'xoauth2handler: access_token "%s expires %s"' % (
             self.oauth2_access_token, self.oauth2_access_token_expires_at))
-        auth_string = 'user=%s\1auth=Bearer %s\1\1'% (
+        auth_string = 'user=%s\1auth=Bearer %s\1\1' % (
             self.username, self.oauth2_access_token)
-        #auth_string = base64.b64encode(auth_string)
-        self.ui.debug('imap', 'xoauth2handler: returning "%s"'% auth_string)
+        # auth_string = base64.b64encode(auth_string)
+        self.ui.debug('imap', 'xoauth2handler: returning "%s"' % auth_string)
         return auth_string
 
     # Perform the next step handling a GSSAPI connection.
@@ -309,12 +310,8 @@ class IMAPServer(object):
             # This is a behavior we got from pykerberos.  First byte is one,
             # first four bytes are preserved (pykerberos calls this a length).
             # Any additional bytes are username.
-            reply = []
-            reply[0:4] = response.message[0:4]
-            reply[0] = '\x01'
-            if self.username:
-                reply[5:] = self.username
-            reply = ''.join(reply)
+            reply = b'\x01' + response.message[1:4]
+            reply += bytes(self.username, 'utf-8')
 
             response = self.gss_vc.wrap(reply, response.encrypted)
             return response.message if response.message else ""
@@ -331,38 +328,39 @@ class IMAPServer(object):
                 imapobj.starttls()
             except imapobj.error as e:
                 raise OfflineImapError("Failed to start "
-                    "TLS connection: %s"% str(e),
-                    OfflineImapError.ERROR.REPO, None, exc_info()[2])
+                                       "TLS connection: %s" % str(e),
+                                       OfflineImapError.ERROR.REPO,
+                                       exc_info()[2])
 
-    ## All __authn_* procedures are helpers that do authentication.
-    ## They are class methods that take one parameter, IMAP object.
-    ##
-    ## Each function should return True if authentication was
-    ## successful and False if authentication wasn't even tried
-    ## for some reason (but not when IMAP has no such authentication
-    ## capability, calling code checks that).
-    ##
-    ## Functions can also raise exceptions; two types are special
-    ## and will be handled by the calling code:
-    ##
-    ## - imapobj.error means that there was some error that
-    ##   comes from imaplib2;
-    ##
-    ## - OfflineImapError means that function detected some
-    ##   problem by itself.
+    # All __authn_* procedures are helpers that do authentication.
+    # They are class methods that take one parameter, IMAP object.
+    #
+    # Each function should return True if authentication was
+    # successful and False if authentication wasn't even tried
+    # for some reason (but not when IMAP has no such authentication
+    # capability, calling code checks that).
+    #
+    # Functions can also raise exceptions; two types are special
+    # and will be handled by the calling code:
+    #
+    # - imapobj.error means that there was some error that
+    #   comes from imaplib2;
+    #
+    # - OfflineImapError means that function detected some
+    #   problem by itself.
 
     def __authn_gssapi(self, imapobj):
         if not have_gss:
             return False
 
         self.connectionlock.acquire()
-        self.gssapi = False
         try:
             imapobj.authenticate('GSSAPI', self.__gsshandler)
-            self.gssapi = True
             return True
+        except imapobj.error:
+            self.gssapi = False
+            raise
         finally:
-            self.gss_vc = None
             self.connectionlock.release()
 
     def __authn_cram_md5(self, imapobj):
@@ -386,8 +384,8 @@ class IMAPServer(object):
         # (per RFC 2595)
         if 'LOGINDISABLED' in imapobj.capabilities:
             raise OfflineImapError("IMAP LOGIN is "
-                "disabled by server.  Need to use SSL?",
-                OfflineImapError.ERROR.REPO)
+                                   "disabled by server.  Need to use SSL?",
+                                   OfflineImapError.ERROR.REPO)
         else:
             self.__loginauth(imapobj)
             return True
@@ -413,11 +411,11 @@ class IMAPServer(object):
         # - tryTLS flag,
         # - check IMAP capability flag.
         auth_methods = {
-          "GSSAPI": (self.__authn_gssapi, False, True),
-          "XOAUTH2": (self.__authn_xoauth2, True, True),
-          "CRAM-MD5": (self.__authn_cram_md5, True, True),
-          "PLAIN": (self.__authn_plain, True, True),
-          "LOGIN": (self.__authn_login, True, False),
+            "GSSAPI": (self.__authn_gssapi, False, True),
+            "XOAUTH2": (self.__authn_xoauth2, True, True),
+            "CRAM-MD5": (self.__authn_cram_md5, True, True),
+            "PLAIN": (self.__authn_plain, True, True),
+            "LOGIN": (self.__authn_login, True, False),
         }
 
         # GSSAPI is tried first by default: we will probably go TLS after it and
@@ -425,7 +423,7 @@ class IMAPServer(object):
         for m in self.authmechs:
             if m not in auth_methods:
                 raise Exception("Bad authentication method %s, "
-                  "please, file OfflineIMAP bug" % m)
+                                "please, file OfflineIMAP bug" % m)
 
             func, tryTLS, check_cap = auth_methods[m]
 
@@ -441,28 +439,28 @@ class IMAPServer(object):
                     continue
 
             tried_to_authn = True
-            self.ui.debug('imap', u'Attempting '
-              '%s authentication'% m)
+            self.ui.debug('imap', 'Attempting '
+                                  '%s authentication' % m)
             try:
                 if func(imapobj):
                     return
             except (imapobj.error, OfflineImapError) as e:
-                self.ui.warn('%s authentication failed: %s'% (m, e))
+                self.ui.warn('%s authentication failed: %s' % (m, e))
                 exc_stack.append((m, e))
 
         if len(exc_stack):
             msg = "\n\t".join([": ".join((x[0], str(x[1]))) for x in exc_stack])
             raise OfflineImapError("All authentication types "
-              "failed:\n\t%s"% msg, OfflineImapError.ERROR.REPO)
+                                   "failed:\n\t%s" % msg, OfflineImapError.ERROR.REPO)
 
         if not tried_to_authn:
             methods = ", ".join([x[5:] for x in
-                [x for x in imapobj.capabilities if x[0:5] == "AUTH="]])
-            raise OfflineImapError(u"Repository %s: no supported "
-              "authentication mechanisms found; configured %s, "
-              "server advertises %s"% (self.repos,
-              ", ".join(self.authmechs), methods),
-              OfflineImapError.ERROR.REPO)
+                                 [x for x in imapobj.capabilities if x[0:5] == "AUTH="]])
+            raise OfflineImapError("Repository %s: no supported "
+                                   "authentication mechanisms found; configured %s, "
+                                   "server advertises %s" % (self.repos,
+                                                             ", ".join(self.authmechs), methods),
+                                   OfflineImapError.ERROR.REPO)
 
     def __verifycert(self, cert, hostname):
         """Verify that cert (in socket.getpeercert() format) matches hostname.
@@ -472,7 +470,7 @@ class IMAPServer(object):
 
         errstr = "CA Cert verifying failed: "
         if not cert:
-            return ('%s no certificate received'% errstr)
+            return '%s no certificate received' % errstr
         dnsname = hostname.lower()
         certnames = []
 
@@ -480,7 +478,7 @@ class IMAPServer(object):
         notafter = cert.get('notAfter')
         if notafter:
             if time.time() >= cert_time_to_seconds(notafter):
-                return '%s certificate expired %s'% (errstr, notafter)
+                return '%s certificate expired %s' % (errstr, notafter)
 
         # First read commonName
         for s in cert.get('subject', []):
@@ -488,7 +486,7 @@ class IMAPServer(object):
             if key == 'commonName':
                 certnames.append(value.lower())
         if len(certnames) == 0:
-            return ('%s no commonName found in certificate'% errstr)
+            return '%s no commonName found in certificate' % errstr
 
         # Then read subjectAltName
         for key, value in cert.get('subjectAltName', []):
@@ -498,10 +496,10 @@ class IMAPServer(object):
         # And finally try to match hostname with one of these names
         for certname in certnames:
             if (certname == dnsname or
-                '.' in dnsname and certname == '*.' + dnsname.split('.', 1)[1]):
+                    '.' in dnsname and certname == '*.' + dnsname.split('.', 1)[1]):
                 return None
 
-        return ('%s no matching domain name found in certificate'% errstr)
+        return '%s no matching domain name found in certificate' % errstr
 
     def acquireconnection(self):
         """Fetches a connection from the pool, making sure to create a new one
@@ -514,7 +512,11 @@ class IMAPServer(object):
         curThread = currentThread()
         imapobj = None
 
-        if len(self.availableconnections): # One is available.
+        imap_debug = 0
+        if 'imap' in self.ui.debuglist:
+            imap_debug = 5
+
+        if len(self.availableconnections):  # One is available.
             # Try to find one that previously belonged to this thread
             # as an optimization.  Start from the back since that's where
             # they're popped on.
@@ -522,17 +524,17 @@ class IMAPServer(object):
                 tryobj = self.availableconnections[i]
                 if self.lastowner[tryobj] == curThread.ident:
                     imapobj = tryobj
-                    del(self.availableconnections[i])
+                    del (self.availableconnections[i])
                     break
             if not imapobj:
                 imapobj = self.availableconnections[0]
-                del(self.availableconnections[0])
+                del (self.availableconnections[0])
             self.assignedconnections.append(imapobj)
             self.lastowner[imapobj] = curThread.ident
             self.connectionlock.release()
             return imapobj
 
-        self.connectionlock.release()   # Release until need to modify data
+        self.connectionlock.release()  # Release until need to modify data
 
         # Must be careful here that if we fail we should bail out gracefully
         # and release locks / threads so that the next attempt can try...
@@ -546,14 +548,15 @@ class IMAPServer(object):
                     imapobj = imaplibutil.IMAP4_Tunnel(
                         self.tunnel,
                         timeout=socket.getdefaulttimeout(),
+                        debug=imap_debug,
                         use_socket=self.proxied_socket,
-                        )
+                    )
                     success = True
                 elif self.usessl:
                     self.ui.connecting(
                         self.repos.getname(), self.hostname, self.port)
-                    self.ui.debug('imap', "%s: level '%s', version '%s'"%
-                        (self.repos.getname(), self.tlslevel, self.sslversion))
+                    self.ui.debug('imap', "%s: level '%s', version '%s'" %
+                                  (self.repos.getname(), self.tlslevel, self.sslversion))
                     imapobj = imaplibutil.WrappedIMAP4_SSL(
                         host=self.hostname,
                         port=self.port,
@@ -562,12 +565,13 @@ class IMAPServer(object):
                         ca_certs=self.sslcacertfile,
                         cert_verify_cb=self.__verifycert,
                         ssl_version=self.sslversion,
+                        debug=imap_debug,
                         timeout=socket.getdefaulttimeout(),
                         fingerprint=self.fingerprint,
                         use_socket=self.proxied_socket,
                         tls_level=self.tlslevel,
                         af=self.af,
-                        )
+                    )
                 else:
                     self.ui.connecting(
                         self.repos.getname(), self.hostname, self.port)
@@ -575,8 +579,14 @@ class IMAPServer(object):
                         self.hostname, self.port,
                         timeout=socket.getdefaulttimeout(),
                         use_socket=self.proxied_socket,
+                        debug=imap_debug,
                         af=self.af,
-                        )
+                    )
+
+                # If 'ID' extension is used by the server, we should use it
+                if 'ID' in imapobj.capabilities:
+                    l_str = '("name" "OfflineIMAP" "version" "{}")'.format(offlineimap.__version__)
+                    imapobj.id(l_str)
 
                 if not self.preauth_tunnel:
                     try:
@@ -594,24 +604,26 @@ class IMAPServer(object):
             # update capabilities after login, e.g. gmail serves different ones
             typ, dat = imapobj.capability()
             if dat != [None]:
-                imapobj.capabilities = tuple(dat[-1].upper().split())
+                # Get the capabilities and convert them to string from bytes
+                s_dat = [x.decode('utf-8') for x in dat[-1].upper().split()]
+                imapobj.capabilities = tuple(s_dat)
 
-            if self.delim == None:
+            if self.delim is None:
                 listres = imapobj.list(self.reference, '""')[1]
-                if listres == [None] or listres == None:
+                if listres == [None] or listres is None:
                     # Some buggy IMAP servers do not respond well to LIST "" ""
                     # Work around them.
                     listres = imapobj.list(self.reference, '"*"')[1]
-                if listres == [None] or listres == None:
+                if listres == [None] or listres is None:
                     # No Folders were returned. This occurs, e.g. if the
                     # 'reference' prefix does not exist on the mail
                     # server. Raise exception.
-                    err = "Server '%s' returned no folders in '%s'"% \
-                        (self.repos.getname(), self.reference)
+                    err = "Server '%s' returned no folders in '%s'" % \
+                          (self.repos.getname(), self.reference)
                     self.ui.warn(err)
                     raise Exception(err)
                 self.delim, self.root = \
-                     imaputil.imapsplit(listres[0])[1:]
+                    imaputil.imapsplit(listres[0])[1:]
                 self.delim = imaputil.dequote(self.delim)
                 self.root = imaputil.dequote(self.root)
 
@@ -628,52 +640,52 @@ class IMAPServer(object):
 
             severity = OfflineImapError.ERROR.REPO
             if type(e) == gaierror:
-                #DNS related errors. Abort Repo sync
-                #TODO: special error msg for e.errno == 2 "Name or service not known"?
-                reason = "Could not resolve name '%s' for repository "\
-                         "'%s'. Make sure you have configured the ser"\
-                         "ver name correctly and that you are online."%\
+                # DNS related errors. Abort Repo sync
+                # TODO: special error msg for e.errno == 2 "Name or service not known"?
+                reason = "Could not resolve name '%s' for repository " \
+                         "'%s'. Make sure you have configured the ser" \
+                         "ver name correctly and that you are online." % \
                          (self.hostname, self.repos)
-                six.reraise(OfflineImapError,
-                            OfflineImapError(reason, severity),
-                            exc_info()[2])
+                raise OfflineImapError(reason, severity, exc_info()[2])
 
             elif isinstance(e, SSLError) and e.errno == errno.EPERM:
                 # SSL unknown protocol error
                 # happens e.g. when connecting via SSL to a non-SSL service
                 if self.port != 993:
-                    reason = "Could not connect via SSL to host '%s' and non-s"\
-                        "tandard ssl port %d configured. Make sure you connect"\
-                        " to the correct port. Got: %s"% (
-                            self.hostname, self.port, e)
+                    reason = "Could not connect via SSL to host '%s' and non-s" \
+                             "tandard ssl port %d configured. Make sure you connect" \
+                             " to the correct port. Got: %s" % (
+                                 self.hostname, self.port, e)
                 else:
-                    reason = "Unknown SSL protocol connecting to host '%s' for "\
-                         "repository '%s'. OpenSSL responded:\n%s"\
-                         % (self.hostname, self.repos, e)
-                six.reraise(OfflineImapError,
-                            OfflineImapError(reason, severity),
-                            exc_info()[2])
+                    reason = "Unknown SSL protocol connecting to host '%s' for " \
+                             "repository '%s'. OpenSSL responded:\n%s" \
+                             % (self.hostname, self.repos, e)
+                raise OfflineImapError(reason, severity, exc_info()[2])
 
-            elif isinstance(e, socket.error) and e.args[0] == errno.ECONNREFUSED:
+            elif isinstance(e, socket.error) and e.args and e.args[0] == errno.ECONNREFUSED:
                 # "Connection refused", can be a non-existing port, or an unauthorized
                 # webproxy (open WLAN?)
-                reason = "Connection to host '%s:%d' for repository '%s' was "\
-                    "refused. Make sure you have the right host and port "\
-                    "configured and that you are actually able to access the "\
-                    "network."% (self.hostname, self.port, self.repos)
-                six.reraise(OfflineImapError,
-                            OfflineImapError(reason, severity),
-                            exc_info()[2])
+                reason = "Connection to host '%s:%d' for repository '%s' was " \
+                         "refused. Make sure you have the right host and port " \
+                         "configured and that you are actually able to access the " \
+                         "network." % (self.hostname, self.port, self.repos)
+                raise OfflineImapError(reason, severity, exc_info()[2])
+
             # Could not acquire connection to the remote;
             # socket.error(last_error) raised
             if str(e)[:24] == "can't open socket; error":
-                six.reraise(OfflineImapError,
-                            OfflineImapError(
-                                "Could not connect to remote server '%s' "
-                                "for repository '%s'. Remote does not answer."%
-                                (self.hostname, self.repos),
-                                OfflineImapError.ERROR.REPO),
-                            exc_info()[2])
+                raise OfflineImapError(
+                    "Could not connect to remote server '%s' "
+                    "for repository '%s'. Remote does not answer." % (self.hostname, self.repos),
+                    OfflineImapError.ERROR.REPO,
+                    exc_info()[2])
+            elif e.args and \
+                    e.args[0][:35] == 'IMAP4 protocol error: socket error:':
+                raise OfflineImapError(
+                    "Could not connect to remote server '{}' "
+                    "for repository '{}'. Connection Refused.".format(
+                        self.hostname, self.repos),
+                    OfflineImapError.ERROR.CRITICAL)
             else:
                 # re-raise all other errors
                 raise
@@ -688,7 +700,7 @@ class IMAPServer(object):
         It's OK if we have maxconnections + 1 or 2 threads, which is what this
         will help us do."""
 
-        self.semaphore.acquire() # Blocking until maxconnections has free slots.
+        self.semaphore.acquire()  # Blocking until maxconnections has free slots.
         self.semaphore.release()
 
     def close(self):
@@ -726,8 +738,8 @@ class IMAPServer(object):
 
             threads = []
             for i in range(numconnections):
-                self.ui.debug('imap', 'keepalive: processing connection %d of %d'%
-                                      (i, numconnections))
+                self.ui.debug('imap', 'keepalive: processing connection %d of %d' %
+                              (i, numconnections))
                 if len(self.idlefolders) > i:
                     # IDLE thread
                     idler = IdleThread(self, self.idlefolders[i])
@@ -749,15 +761,15 @@ class IMAPServer(object):
         self.ui.debug('imap', 'keepalive: event is set; exiting')
         return
 
-
     def releaseconnection(self, connection, drop_conn=False):
         """Releases a connection, returning it to the pool.
 
+        :param connection: Connection object
         :param drop_conn: If True, the connection will be released and
            not be reused. This can be used to indicate broken connections."""
 
         if connection is None:
-            return # Noop on bad connection.
+            return  # Noop on bad connection.
 
         self.connectionlock.acquire()
         self.assignedconnections.remove(connection)
@@ -770,7 +782,7 @@ class IMAPServer(object):
         self.semaphore.release()
 
 
-class IdleThread(object):
+class IdleThread:
     def __init__(self, parent, folder=None):
         """If invoked without 'folder', perform a NOOP and wait for
         self.stop() to be called. If invoked with folder, switch to IDLE
@@ -784,7 +796,7 @@ class IdleThread(object):
             self.thread = Thread(target=self.noop)
         else:
             self.thread = Thread(target=self.__idle)
-        self.thread.setDaemon(1)
+        self.thread.setDaemon(True)
 
     def start(self):
         self.thread.start()
@@ -806,21 +818,19 @@ class IdleThread(object):
         try:
             imapobj.noop()
         except imapobj.abort:
-            self.ui.warn('Attempting NOOP on dropped connection %s'%
-                imapobj.identifier)
+            self.ui.warn('Attempting NOOP on dropped connection %s' %
+                         imapobj.identifier)
             self.parent.releaseconnection(imapobj, True)
             imapobj = None
         finally:
             if imapobj:
                 self.parent.releaseconnection(imapobj)
-                self.stop_sig.wait() # wait until we are supposed to quit
+                self.stop_sig.wait()  # wait until we are supposed to quit
 
     def __dosync(self):
         remoterepos = self.parent.repos
         account = remoterepos.account
-        localrepos = account.localrepos
         remoterepos = account.remoterepos
-        statusrepos = account.statusrepos
         remotefolder = remoterepos.getfolder(self.folder, decode=False)
 
         hook = account.getconf('presynchook', '')
@@ -830,7 +840,7 @@ class IdleThread(object):
         account.callhook(hook)
 
         ui = getglobalui()
-        ui.unregisterthread(currentThread()) #syncfolder registered the thread
+        ui.unregisterthread(currentThread())  # syncfolder registered the thread
 
     def __idle(self):
         """Invoke IDLE mode until timeout or self.stop() is invoked."""
@@ -847,7 +857,7 @@ class IdleThread(object):
             if exc_data is None and not self.stop_sig.isSet():
                 # No Exception, and we are not supposed to stop:
                 self.needsync = True
-            self.stop_sig.set() # Continue to sync.
+            self.stop_sig.set()  # Continue to sync.
 
         def noop(imapobj):
             """Factorize the noop code."""
@@ -856,17 +866,16 @@ class IdleThread(object):
                 # End IDLE mode with noop, imapobj can point to a dropped conn.
                 imapobj.noop()
             except imapobj.abort:
-                self.ui.warn('Attempting NOOP on dropped connection %s'%
-                    imapobj.identifier)
+                self.ui.warn('Attempting NOOP on dropped connection %s' %
+                             imapobj.identifier)
                 self.parent.releaseconnection(imapobj, True)
             else:
                 self.parent.releaseconnection(imapobj)
 
-
         while not self.stop_sig.isSet():
             self.needsync = False
 
-            success = False # Successfully selected FOLDER?
+            success = False  # Successfully selected FOLDER?
             while not success:
                 imapobj = self.parent.acquireconnection()
                 try:
@@ -888,9 +897,9 @@ class IdleThread(object):
                 imapobj.idle(callback=callback)
             else:
                 self.ui.warn("IMAP IDLE not supported on server '%s'."
-                    "Sleep until next refresh cycle."% imapobj.identifier)
-                noop(imapobj) #XXX: why?
-            self.stop_sig.wait() # self.stop() or IDLE callback are invoked.
+                             "Sleep until next refresh cycle." % imapobj.identifier)
+                noop(imapobj)  # XXX: why?
+            self.stop_sig.wait()  # self.stop() or IDLE callback are invoked.
             noop(imapobj)
 
             if self.needsync:
