@@ -1,5 +1,5 @@
 # OfflineIMAP initialization code
-# Copyright (C) 2002-2016 John Goerzen & contributors
+# Copyright (C) 2002-2017 John Goerzen & contributors
 #
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -26,7 +26,8 @@ import collections
 from optparse import OptionParser
 
 import offlineimap
-import offlineimap.virtual_imaplib2 as imaplib
+from offlineimap.utils.distro_utils import get_os_name
+import imaplib2 as imaplib
 
 # Ensure that `ui` gets loaded before `threadutil` in order to
 # break the circular dependency between `threadutil` and `Curses`.
@@ -38,7 +39,6 @@ from offlineimap.utils import stacktrace
 from offlineimap.repository import Repository
 from offlineimap.folder.IMAP import MSGCOPY_NAMESPACE
 
-
 ACCOUNT_LIMITED_THREAD_NAME = 'MAX_ACCOUNTS'
 PYTHON_VERSION = sys.version.split(' ')[0]
 
@@ -46,24 +46,24 @@ PYTHON_VERSION = sys.version.split(' ')[0]
 def syncitall(list_accounts, config):
     """The target when in multithreading mode for running accounts threads."""
 
-    threads = threadutil.accountThreads() # The collection of accounts threads.
+    threads = threadutil.accountThreads()  # The collection of accounts threads.
     for accountname in list_accounts:
         # Start a new thread per account and store it in the collection.
         account = accounts.SyncableAccount(config, accountname)
         thread = threadutil.InstanceLimitedThread(
             ACCOUNT_LIMITED_THREAD_NAME,
-            target = account.syncrunner,
-            name = "Account sync %s"% accountname
-            )
+            target=account.syncrunner,
+            name="Account sync %s" % accountname
+        )
         thread.setDaemon(True)
         # The add() method expects a started thread.
         thread.start()
         threads.add(thread)
     # Wait for the threads to finish.
-    threads.wait() # Blocks until all accounts are processed.
+    threads.wait()  # Blocks until all accounts are processed.
 
 
-class OfflineImap(object):
+class OfflineImap:
     """The main class that encapsulates the high level use of OfflineImap.
 
     To invoke OfflineImap you would call it with::
@@ -72,6 +72,26 @@ class OfflineImap(object):
       oi.run()
     """
 
+    def get_env_info(self):
+        # Transitional code between imaplib2 versions
+        try:
+            # imaplib2, previous versions, based on Python 2.x
+            l_imaplib_version = imaplib.__version__
+        except AttributeError:
+            # New imaplib2, version >= 3.06
+            l_imaplib_version = imaplib.version()
+        except:
+            # This should not happen
+            l_imaplib_version = " Unknown"
+
+        info = "imaplib2 v%s, Python v%s" % (l_imaplib_version, PYTHON_VERSION)
+        try:
+            import ssl
+            info = "%s, %s" % (info, ssl.OPENSSL_VERSION)
+        except:
+            pass
+        return info
+
     def run(self):
         """Parse the commandline and invoke everything"""
         # next line also sets self.config and self.ui
@@ -92,105 +112,104 @@ class OfflineImap(object):
     def __parse_cmd_options(self):
         parser = OptionParser(
             version=offlineimap.__version__,
-            description="%s.\n\n%s"% (offlineimap.__copyright__,
-                offlineimap.__license__)
-            )
+            description="%s.\n\n%s" % (offlineimap.__copyright__,
+                                       offlineimap.__license__)
+        )
 
         parser.add_option("-V",
-                  action="store_true", dest="version",
-                  default=False,
-                  help="show full version infos")
+                          action="store_true", dest="version",
+                          default=False,
+                          help="show full version infos")
 
         parser.add_option("--dry-run",
-                  action="store_true", dest="dryrun",
-                  default=False,
-                  help="dry run mode")
+                          action="store_true", dest="dryrun",
+                          default=False,
+                          help="dry run mode")
 
         parser.add_option("--info",
-                  action="store_true", dest="diagnostics",
-                  default=False,
-                  help="output information on the configured email repositories")
+                          action="store_true", dest="diagnostics",
+                          default=False,
+                          help="output information on the configured email repositories")
 
         parser.add_option("-1",
-                  action="store_true", dest="singlethreading",
-                  default=False,
-                  help="(the number one) disable all multithreading operations")
+                          action="store_true", dest="singlethreading",
+                          default=False,
+                          help="(the number one) disable all multithreading operations")
 
         parser.add_option("-P", dest="profiledir", metavar="DIR",
-                  help="sets OfflineIMAP into profile mode.")
+                          help="sets OfflineIMAP into profile mode.")
 
         parser.add_option("-a", dest="accounts",
-                  metavar="account1[,account2[,...]]",
-                  help="list of accounts to sync")
+                          metavar="account1[,account2[,...]]",
+                          help="list of accounts to sync")
 
         parser.add_option("-c", dest="configfile", metavar="FILE",
-                  default=None,
-                  help="specifies a configuration file to use")
+                          default=None,
+                          help="specifies a configuration file to use")
 
         parser.add_option("-d", dest="debugtype",
-                  metavar="type1[,type2[,...]]",
-                  help="enables debugging for OfflineIMAP "
-                  " (types: imap, maildir, thread)")
+                          metavar="type1[,type2[,...]]",
+                          help="enables debugging for OfflineIMAP "
+                               " (types: imap, maildir, thread)")
 
         parser.add_option("-l", dest="logfile", metavar="FILE",
-                  help="log to FILE")
+                          help="log to FILE")
 
         parser.add_option("-s",
-                  action="store_true", dest="syslog",
-                  default=False,
-                  help="log to syslog")
+                          action="store_true", dest="syslog",
+                          default=False,
+                          help="log to syslog")
 
         parser.add_option("-f", dest="folders",
-                  metavar="folder1[,folder2[,...]]",
-                  help="only sync the specified folders")
+                          metavar="folder1[,folder2[,...]]",
+                          help="only sync the specified folders")
 
         parser.add_option("-k", dest="configoverride",
-                  action="append",
-                  metavar="[section:]option=value",
-                  help="override configuration file option")
+                          action="append",
+                          metavar="[section:]option=value",
+                          help="override configuration file option")
 
         parser.add_option("-o",
-                  action="store_true", dest="runonce",
-                  default=False,
-                  help="run only once (ignore autorefresh)")
+                          action="store_true", dest="runonce",
+                          default=False,
+                          help="run only once (ignore autorefresh)")
 
         parser.add_option("-q",
-                  action="store_true", dest="quick",
-                  default=False,
-                  help="run only quick synchronizations (don't update flags)")
+                          action="store_true", dest="quick",
+                          default=False,
+                          help="run only quick synchronizations (don't update flags)")
 
         parser.add_option("-u", dest="interface",
-                  help="specifies an alternative user interface"
-                  " (quiet, basic, syslog, ttyui, blinkenlights, machineui)")
+                          help="specifies an alternative user interface"
+                               " (quiet, basic, syslog, ttyui, blinkenlights, machineui)")
 
         parser.add_option("--delete-folder", dest="deletefolder",
-                  default=None,
-                  metavar="FOLDERNAME",
-                  help="Delete a folder (on the remote repository)")
+                          default=None,
+                          metavar="FOLDERNAME",
+                          help="Delete a folder (on the remote repository)")
 
         parser.add_option("--migrate-fmd5-using-nametrans",
-                  action="store_true", dest="migrate_fmd5", default=False,
-                  help="migrate FMD5 hashes from versions prior to 6.3.5")
+                          action="store_true", dest="migrate_fmd5", default=False,
+                          help="migrate FMD5 hashes from versions prior to 6.3.5")
 
         parser.add_option("--mbnames-prune",
-                  action="store_true", dest="mbnames_prune", default=False,
-                  help="remove mbnames entries for accounts not in accounts")
+                          action="store_true", dest="mbnames_prune", default=False,
+                          help="remove mbnames entries for accounts not in accounts")
 
         (options, args) = parser.parse_args()
         glob.set_options(options)
 
         if options.version:
-            print("offlineimap v%s, imaplib2 v%s (%s), Python v%s"% (
-                  offlineimap.__version__, imaplib.__version__, imaplib.DESC,
-                  PYTHON_VERSION)
-            )
+            print(("offlineimap v%s, %s" % (
+                offlineimap.__version__, self.get_env_info())
+                   ))
             sys.exit(0)
 
         # Read in configuration file.
         if not options.configfile:
             # Try XDG location, then fall back to ~/.offlineimaprc
             xdg_var = 'XDG_CONFIG_HOME'
-            if not xdg_var in os.environ or not os.environ[xdg_var]:
+            if xdg_var not in os.environ or not os.environ[xdg_var]:
                 xdg_home = os.path.expanduser('~/.config')
             else:
                 xdg_home = os.environ[xdg_var]
@@ -204,7 +223,7 @@ class OfflineImap(object):
         config = CustomConfigParser()
         if not os.path.exists(configfilename):
             # TODO, initialize and make use of chosen ui for logging
-            logging.error(" *** Config file '%s' does not exist; aborting!"%
+            logging.error(" *** Config file '%s' does not exist; aborting!" %
                           configfilename)
             sys.exit(1)
         config.read(configfilename)
@@ -213,17 +232,17 @@ class OfflineImap(object):
         if options.profiledir:
             if not options.singlethreading:
                 # TODO, make use of chosen ui for logging
-                logging.warn("Profile mode: Forcing to singlethreaded.")
+                logging.warning("Profile mode: Forcing to singlethreaded.")
                 options.singlethreading = True
             if os.path.exists(options.profiledir):
                 # TODO, make use of chosen ui for logging
-                logging.warn("Profile mode: Directory '%s' already exists!"%
-                             options.profiledir)
+                logging.warning("Profile mode: Directory '%s' already exists!" %
+                                options.profiledir)
             else:
                 os.mkdir(options.profiledir)
             # TODO, make use of chosen ui for logging
-            logging.warn("Profile mode: Potentially large data will be "
-                         "created in '%s'"% options.profiledir)
+            logging.warning("Profile mode: Potentially large data will be "
+                            "created in '%s'" % options.profiledir)
 
         # Override a config value.
         if options.configoverride:
@@ -238,28 +257,28 @@ class OfflineImap(object):
 
         # Which ui to use? CLI option overrides config file.
         ui_type = config.getdefault('general', 'ui', 'ttyui')
-        if options.interface != None:
+        if options.interface is not None:
             ui_type = options.interface
         if '.' in ui_type:
             # Transform Curses.Blinkenlights -> Blinkenlights.
             ui_type = ui_type.split('.')[-1]
             # TODO, make use of chosen ui for logging
             logging.warning('Using old interface name, consider using one '
-                            'of %s'% ', '.join(UI_LIST.keys()))
+                            'of %s' % ', '.join(list(UI_LIST.keys())))
         if options.diagnostics:
-            ui_type = 'ttyui' # Enforce this UI for --info.
+            ui_type = 'ttyui'  # Enforce this UI for --info.
 
         # dry-run? Set [general]dry-run=True.
         if options.dryrun:
-            dryrun = config.set('general', 'dry-run', 'True')
+            config.set('general', 'dry-run', 'True')
         config.set_if_not_exists('general', 'dry-run', 'False')
 
         try:
             # Create the ui class.
             self.ui = UI_LIST[ui_type.lower()](config)
         except KeyError:
-            logging.error("UI '%s' does not exist, choose one of: %s"%
-                (ui_type, ', '.join(UI_LIST.keys())))
+            logging.error("UI '%s' does not exist, choose one of: %s" %
+                          (ui_type, ', '.join(list(UI_LIST.keys()))))
             sys.exit(1)
         setglobalui(self.ui)
 
@@ -273,13 +292,14 @@ class OfflineImap(object):
 
         # Welcome blurb.
         self.ui.init_banner()
+        self.ui.info(self.get_env_info())
 
         if options.debugtype:
             self.ui.logger.setLevel(logging.DEBUG)
             if options.debugtype.lower() == 'all':
                 options.debugtype = 'imap,maildir,thread'
             # Force single threading?
-            if not ('thread' in options.debugtype.split(',') \
+            if not ('thread' in options.debugtype.split(',')
                     and not options.singlethreading):
                 self.ui._msg("Debug mode: Forcing to singlethreaded.")
                 options.singlethreading = True
@@ -288,8 +308,6 @@ class OfflineImap(object):
             for dtype in debugtypes:
                 dtype = dtype.strip()
                 self.ui.add_debug(dtype)
-                if dtype.lower() == u'imap':
-                    imaplib.Debug = 5
 
         if options.runonce:
             # Must kill the possible default option.
@@ -306,12 +324,12 @@ class OfflineImap(object):
         # Custom folder list specified?
         if options.folders:
             foldernames = options.folders.split(",")
-            folderfilter = "lambda f: f in %s"% foldernames
+            folderfilter = "lambda f: f in %s" % foldernames
             folderincludes = "[]"
             for accountname in accounts.getaccountlist(config):
                 account_section = 'Account ' + accountname
                 remote_repo_section = 'Repository ' + \
-                    config.get(account_section, 'remoterepository')
+                                      config.get(account_section, 'remoterepository')
                 config.set(remote_repo_section, "folderfilter", folderfilter)
                 config.set(remote_repo_section, "folderincludes",
                            folderincludes)
@@ -326,13 +344,13 @@ class OfflineImap(object):
         threadutil.initInstanceLimit(
             ACCOUNT_LIMITED_THREAD_NAME,
             config.getdefaultint('general', 'maxsyncaccounts', 1)
-            )
+        )
 
         for reposname in config.getsectionlist('Repository'):
             # Limit the number of threads. Limitation on usage is handled at the
             # imapserver level.
             for namespace in [accounts.FOLDER_NAMESPACE + reposname,
-                                 MSGCOPY_NAMESPACE + reposname]:
+                              MSGCOPY_NAMESPACE + reposname]:
                 if options.singlethreading:
                     threadutil.initInstanceLimit(namespace, 1)
                 else:
@@ -341,9 +359,9 @@ class OfflineImap(object):
                         config.getdefaultint(
                             'Repository ' + reposname,
                             'maxconnections', 2)
-                        )
+                    )
         self.config = config
-        return (options, args)
+        return options, args
 
     def __dumpstacks(self, context=1, sighandler_deep=2):
         """ Signal handler: dump a stack trace for each existing thread."""
@@ -354,10 +372,10 @@ class OfflineImap(object):
             d = collections.defaultdict(lambda: 0)
             for v in l:
                 d[tuple(v)] += 1
-            return list((k, v) for k, v in d.items())
+            return list((k, v) for k, v in list(d.items()))
 
         stack_displays = []
-        for threadId, stack in sys._current_frames().items():
+        for threadId, stack in list(sys._current_frames().items()):
             stack_display = []
             for filename, lineno, name, line in traceback.extract_stack(stack):
                 stack_display.append('  File: "%s", line %d, in %s'
@@ -378,7 +396,7 @@ class OfflineImap(object):
             self.ui.debug('thread', msg % (times, '\n'.join(stack[- (context * 2):])))
 
         self.ui.debug('thread', "Dumped a total of %d Threads." %
-                      len(sys._current_frames().keys()))
+                      len(list(sys._current_frames().keys())))
 
     def _get_activeaccounts(self, options):
         activeaccounts = []
@@ -394,9 +412,9 @@ class OfflineImap(object):
             if accountname in allaccounts:
                 activeaccounts.append(accountname)
             else:
-                errormsg = "Valid accounts are: %s"% (
+                errormsg = "Valid accounts are: %s" % (
                     ", ".join(allaccounts))
-                self.ui.error("The account '%s' does not exist"% accountname)
+                self.ui.error("The account '%s' does not exist" % accountname)
 
         if len(activeaccounts) < 1:
             errormsg = "No accounts are defined!"
@@ -422,8 +440,9 @@ class OfflineImap(object):
                 accounts.Account.set_abort_event(self.config, 2)
             elif sig in (signal.SIGTERM, signal.SIGINT, signal.SIGHUP):
                 # tell each account to ABORT ASAP (ctrl-c)
-                getglobalui().warn("Terminating NOW (this may "\
-                                   "take a few seconds)...")
+                getglobalui().warn("Preparing to shutdown after sync (this may "
+                                   "take some time), press CTRL-C three "
+                                   "times to shutdown immediately")
                 accounts.Account.set_abort_event(self.config, 3)
                 if 'thread' in self.ui.debuglist:
                     self.__dumpstacks(5)
@@ -439,13 +458,16 @@ class OfflineImap(object):
 
         try:
             self.num_sigterm = 0
-            signal.signal(signal.SIGHUP, sig_handler)
-            signal.signal(signal.SIGUSR1, sig_handler)
-            signal.signal(signal.SIGUSR2, sig_handler)
-            signal.signal(signal.SIGABRT, sig_handler)
-            signal.signal(signal.SIGTERM, sig_handler)
-            signal.signal(signal.SIGINT, sig_handler)
-            signal.signal(signal.SIGQUIT, sig_handler)
+
+            # We cannot use signals in Windows
+            if get_os_name() != 'windows':
+                signal.signal(signal.SIGHUP, sig_handler)
+                signal.signal(signal.SIGUSR1, sig_handler)
+                signal.signal(signal.SIGUSR2, sig_handler)
+                signal.signal(signal.SIGABRT, sig_handler)
+                signal.signal(signal.SIGTERM, sig_handler)
+                signal.signal(signal.SIGINT, sig_handler)
+                signal.signal(signal.SIGQUIT, sig_handler)
 
             # Various initializations that need to be performed:
             activeaccounts = self._get_activeaccounts(options)
@@ -460,7 +482,7 @@ class OfflineImap(object):
                     target=syncitall,
                     name='Sync Runner',
                     args=(activeaccounts, self.config,)
-                    )
+                )
                 # Special exit message for the monitor to stop looping.
                 t.exit_message = threadutil.STOP_MONITOR
                 t.start()
@@ -470,7 +492,7 @@ class OfflineImap(object):
             mbnames.write()
             self.ui.terminate()
             return 0
-        except (SystemExit):
+        except SystemExit:
             raise
         except Exception as e:
             self.ui.error(e)
@@ -485,7 +507,7 @@ class OfflineImap(object):
         for accountname in list_accounts:
             account = accounts.SyncableAccount(self.config, accountname)
             threading.currentThread().name = \
-                    "Account sync %s"% account.getname()
+                "Account sync %s" % account.getname()
             if not profiledir:
                 account.syncrunner()
             # Profile mode.
@@ -502,10 +524,10 @@ class OfflineImap(object):
                 from datetime import datetime
                 dt = datetime.now().strftime('%Y%m%d%H%M%S')
                 prof.dump_stats(os.path.join(
-                    profiledir, "%s_%s.prof"% (dt, account.getname())))
+                    profiledir, "%s_%s.prof" % (dt, account.getname())))
 
     def __serverdiagnostics(self, options):
-        self.ui.info("  imaplib2: %s (%s)"% (imaplib.__version__, imaplib.DESC))
+        self.ui.info("  imaplib2: %s" % imaplib.__version__)
         for accountname in self._get_activeaccounts(options):
             account = accounts.Account(self.config, accountname)
             account.serverdiagnostics()
